#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
/*
 * ctrlpacket.c
 *
 * PPTP control packet reading/writing/formatting methods
 *
 *
 * Matthew Ramsay <matthewr@moreton.com.au>
 *
 */

#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "pptpdebug.h"
#include "pptpdefs.h"
#include "ctrlpacket.h"


// local function prototypes
ssize_t read_pptp_header(int clientFd, unsigned char *packet, int *ctrl_message_type);
int deal_start_ctrl_conn(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size);
int deal_out_call(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size);
int deal_echo(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size);
int deal_call_clr(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size);
//int make_in_call_rqst_pckt(int clientFd, unsigned char *rply_packet, int *rply_size);



/*
 * read_pptp_packet
 *
 * Sees if a packet can be read and if so what type of packet it is.. it then calls the
 * appropriate method to work out the details of the packet and to form a reply packet.
 *
 * args: clientFd (IN) - client socket to read from
 *	 packet (OUT) - storage for the packet we read from the client
 *	 rply_packet (OUT) - storage for the reply packet we need to send to the client
 *	 rply_size (OUT) - the size of the reply packet
 *
 * retn: PPTP control message type of the packet, or -1 on failure
 *
 */
int read_pptp_packet(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size, int pid) {

	size_t bytes_read;
	int pptp_ctrl_type;

	
 	// read a packets header
	if((bytes_read = read_pptp_header(clientFd, packet, &pptp_ctrl_type)) == -1) {
	
		// something went wrong
		pptpd_debug("CTRL: couldn't read packet header", CRITICAL);
		return -1;

	}

	
	switch(pptp_ctrl_type) {
	
	case START_CTRL_CONN_RQST: // the packet was a start control connection request
	
		deal_start_ctrl_conn(clientFd, packet, rply_packet, rply_size);
		break;

	case OUT_CALL_RQST: // the packet was an out call request
			
		deal_out_call(clientFd, packet, rply_packet, rply_size);
		kill(pid, SIGUSR1);
		break;

	case ECHO_RQST: // the packet was an echo request
			
		deal_echo(clientFd, packet, rply_packet, rply_size);
		break;
		
	case CALL_CLR_RQST: // the packet was an echo request
			
		deal_call_clr(clientFd, packet, rply_packet, rply_size);
		break;
		
	default:
		
		pptpd_debug("CTRL: I got a pptp packet.. but I don't know what to do with it!", CRITICAL);
			
	}
	
	
	return pptp_ctrl_type;

}


/*
 * send_pptp_packet
 *
 * Sends a PPTP packet
 *
 *
 */
size_t send_pptp_packet(int clientFd, unsigned char *packet, size_t packet_size) {

	int i;
	size_t bytes_written;
        char debug_msg[MAX_DEBUG_MSG_SIZE];
	
	if((bytes_written = write(clientFd, packet, packet_size)) == -1) {
	
		// oops.. couldn't write
		pptpd_debug("CTRL: couldn't write packet to client", CRITICAL);
		
		return -1;
		
	} else {

		// debug
		sprintf(debug_msg, "CTRL: I wrote %d bytes to the client", packet_size);
		pptpd_debug(debug_msg, VERBOSE);
		pptpd_byte_debug("write", packet, packet_size);
		pptpd_debug("CTRL: sent packet to client", VERBOSE);
		////////////////////////////////////////////////////////
				
		return bytes_written;
		
	}
	
}


///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
// local functions only below
///////////////////////////////////////////////////////////////////////////

/*
 * read_pptp_header
 *
 * Reads a packet storing the packet in "packet" and the PPTP control type
 * in "pptp_ctrl_type" and returns the number of bytes read if all went well
 *
 * args: clientFd (IN) - client we r receiving the packet from
 *	 packet (OUT) - where we will store the packet we get
 *	 pptp_ctrl_type (OUT) - PPTP control type of the packet we got
 * retn: bytes read on success, -1 on failure
 */
ssize_t read_pptp_header(int clientFd, unsigned char *packet,
				int *pptp_ctrl_type) {
	
	ssize_t bytes_read;
	size_t bytes_to_read = PPTP_MAX_CTRL_PCKT_SIZE;
	
	char *debug_msg = malloc(MAX_DEBUG_MSG_SIZE);
	
	struct pptp_header *header;
	
	
	if((bytes_read = read(clientFd, packet, bytes_to_read)) == -1) {
	
		// error reading data
		pptpd_debug("CTRL: couldn't read anything..", CRITICAL);
		return -1;
		
	} else {
	
		header = (struct pptp_header *)	packet;
		
		// check to see if what we got as a magic cookie
		if(ntohl(header->magic) != PPTP_MAGIC_COOKIE) {
			
			pptpd_debug("CTRL: not a pptp message", CRITICAL);
			return -1;
			
		} else { // woohoo! looks like we got a valid packet

			*pptp_ctrl_type = (int) (ntohs(header->ctrl_type));
			
			sprintf(debug_msg, "CTRL: I got a valid PPTP packet with control type %d", *pptp_ctrl_type);
			pptpd_debug(debug_msg, VERBOSE);
			
			pptpd_byte_debug("packet", packet, bytes_read);
			
			return bytes_read;
						
		}
			
	}

}


/*
 * deal_start_ctrl_conn
 *
 *
 * retn: 0 on success
 *
 */
int deal_start_ctrl_conn(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size) {

	int i;
	
       	struct pptp_header rply_header;
       	
	struct pptp_start_ctrl_conn_rqst *start_ctrl_conn_rqst;
	struct pptp_start_ctrl_conn_rply start_ctrl_conn_rply;

	unsigned char pptpd_hostname[MAX_HOSTNAME_SIZE] = PPTP_HOSTNAME;
	unsigned char pptpd_vendor[MAX_VENDOR_SIZE] = PPTP_VENDOR;

        char debug_msg[MAX_DEBUG_MSG_SIZE]; // debugging
	
		
	start_ctrl_conn_rqst = (struct pptp_start_ctrl_conn_rqst *) packet;
				
	////////////////////////// DEBUGGING //////////////////////////////
	pptpd_byte_debug("Hostname sent to me", start_ctrl_conn_rqst->hostname, MAX_HOSTNAME_SIZE);
	pptpd_byte_debug("Vendor name sent to me", start_ctrl_conn_rqst->vendor, MAX_VENDOR_SIZE);
	///////////////////////////////////////////////////////////////////
	
	// form a reply
		
	rply_header.length = htons(0x009c);
	rply_header.pptp_type = htons(PPTP_CTRL_MESSAGE);
	rply_header.magic = htonl(PPTP_MAGIC_COOKIE);
	rply_header.ctrl_type = htons(START_CTRL_CONN_RPLY);
	rply_header.reserved0 = htons(0x0000);
		
	start_ctrl_conn_rply.header = rply_header;
	start_ctrl_conn_rply.version = htons(PPTP_VERSION);
	start_ctrl_conn_rply.result_code = 0x01;
	start_ctrl_conn_rply.error_code = 0x00;
	start_ctrl_conn_rply.framing_cap = htonl(SYNCHRONOUS_FRAMING);
	start_ctrl_conn_rply.bearer_cap = htonl(DIGITAL_ACCESS);
	start_ctrl_conn_rply.max_channels = htons(0x0001);
	start_ctrl_conn_rply.firmware_rev = htons(PPTP_FIRMWARE_VERSION);
	memcpy(start_ctrl_conn_rply.hostname, pptpd_hostname, MAX_HOSTNAME_SIZE);
	memcpy(start_ctrl_conn_rply.vendor, pptpd_vendor, MAX_VENDOR_SIZE);
		
	*rply_size = sizeof(start_ctrl_conn_rply);
	memcpy(rply_packet, &start_ctrl_conn_rply, *rply_size);
		
	pptpd_debug("CTRL: I've made a suitable START_CTRL_CONN_RPLY..", VERBOSE);
	
	return 0; // maybe one day i'll put some error checking in
	
}


/*
 * deal_out_call
 *
 *
 * retn: 0 on success
 *
 */
int deal_out_call(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size) {

	int i;
	
       	struct pptp_header rply_header;
       	
	struct pptp_out_call_rqst *out_call_rqst;
	struct pptp_out_call_rply out_call_rply;

	unsigned char callID[2];
	unsigned char maxBPS[4];
	unsigned char packetRecvSize[2];
	unsigned char packetProcessDelay[2];

	size_t packet_size; // size of the packet being referred to

        char debug_msg[MAX_DEBUG_MSG_SIZE]; // debugging

        		
	out_call_rqst = (struct pptp_out_call_rqst *) packet;
		
	memcpy(callID, &out_call_rqst->call_id, 2);
	memcpy(maxBPS, &out_call_rqst->max_bps, 4);
	memcpy(packetRecvSize, &out_call_rqst->pckt_recv_size, 2);
	memcpy(packetProcessDelay, &out_call_rqst->pckt_delay, 2);

			
	//////////////////// DEBUGGING //////////////////////////////
	pptpd_byte_debug("packet delay", packetProcessDelay, 2);
	pptpd_byte_debug("packet recv size", packetRecvSize, 2);
	pptpd_byte_debug("call ID", callID, 2);
	pptpd_byte_debug("max BPS", maxBPS, 4);

	packet_size = sizeof(out_call_rqst);
	pptpd_byte_debug("entire packet", packet, packet_size);
	////////////////////////////////////////////////////////////////
		
	// make a reply
		
	rply_header.length = htons(0x0020);
	rply_header.pptp_type = htons(PPTP_CTRL_MESSAGE);
	rply_header.magic = htonl(PPTP_MAGIC_COOKIE);
	rply_header.ctrl_type = htons(OUT_CALL_RPLY);
	rply_header.reserved0 = htons(0x0000);
		
	out_call_rply.header = rply_header;
	
	// @FIXME
	// this one was causing problems.. it has to match call id in pptpserver.c
	// i think.. or else..
	out_call_rply.call_id = htons(0x0000); // call id.. made up number for now...
	
	//out_call_rply.call_id_peer = *callID;
	memcpy(&out_call_rply.call_id_peer, callID, 2);
	
	out_call_rply.result_code = 0x01;
	out_call_rply.error_code = 0x00;
	out_call_rply.cause_code = 0x00;
	
	//out_call_rply.speed = *maxBPS; // lets go as fast as the client can go
	memcpy(&out_call_rply.speed, maxBPS, 4);	
	
	out_call_rply.pckt_recv_size = htons(0x0000);
	out_call_rply.pckt_delay = htons(0x0000);
	out_call_rply.channel_id = htonl(0x00000000);
		
	*rply_size = sizeof(out_call_rply);
	memcpy(rply_packet, &out_call_rply, *rply_size);
		
	pptpd_debug("CTRL: I've made a suitable OUT_CALL_RPLY..", VERBOSE);
						
	return 0;
		
}
		
		
/*
 * deal_echo
 *
 *
 * retn: 0 on success
 *
 */		
int deal_echo(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size) {

	int i;
	
       	struct pptp_header rply_header;
       	
	struct pptp_echo_rqst *echo_rqst;
	struct pptp_echo_rply echo_rply;	

	unsigned char echoID[4];
	
        char debug_msg[MAX_DEBUG_MSG_SIZE]; // debugging

	echo_rqst = (struct pptp_echo_rqst *) packet;
		
	memcpy(echoID, &echo_rqst->identifier, 4);
		
	////////////////////////// DEBUGGING //////////////////////////////
	pptpd_byte_debug("echo id", echoID, 4);
	///////////////////////////////////////////////////////////////////

	// form a reply
		
	rply_header.length = htons(0x0014);
	rply_header.pptp_type = htons(PPTP_CTRL_MESSAGE);
	rply_header.magic = htonl(PPTP_MAGIC_COOKIE);
	rply_header.ctrl_type = htons(ECHO_RPLY);
	rply_header.reserved0 = htons(0x0000);
		
	echo_rply.header = rply_header;
	//echo_rply.identifier = htonl(0x00000001);//*echoID; // copy from id echo_rqst
	memcpy(&echo_rply.identifier, echoID, 4);
	
	echo_rply.result_code = 0x01;
	echo_rply.error_code = 0x00;
	echo_rply.reserved1 = htons(0x0000);
		
	*rply_size = sizeof(echo_rply);
	memcpy(rply_packet, &echo_rply, *rply_size);
		
	pptpd_debug("CTRL: I've made a suitable ECHO_RPLY..", VERBOSE);
		
	return 0;
	
}


/*
 * deal_call_clr
 *
 *
 * retn: 0 on success
 *
 */		
int deal_call_clr(int clientFd, unsigned char *packet, unsigned char *rply_packet, int *rply_size) {

	int i;
	
       	struct pptp_header rply_header;
       	
	struct pptp_call_clr_rqst *call_clr_rqst;
	struct pptp_call_disconn_ntfy call_disconn_ntfy;

	unsigned char callID[2];
	
	unsigned char callStats[128] = {0};
	
        char debug_msg[MAX_DEBUG_MSG_SIZE]; // debugging

        		
	call_clr_rqst = (struct pptp_call_clr_rqst *) packet;
	
	memcpy(callID, &call_clr_rqst->call_id, 2);

			
	////////////////////////// DEBUGGING //////////////////////////////
	pptpd_byte_debug("call ID", callID, 2);
	///////////////////////////////////////////////////////////////////

	// form a reply
		
	rply_header.length = htons(0x0094);
	rply_header.pptp_type = htons(PPTP_CTRL_MESSAGE);
	rply_header.magic = htonl(PPTP_MAGIC_COOKIE);
	rply_header.ctrl_type = htons(CALL_DISCONN_NTFY);
	rply_header.reserved0 = htons(0x0000);
		
	call_disconn_ntfy.header = rply_header;
	call_disconn_ntfy.call_id = htons(0x0000); // for now (as in out_call_rply) out call id is 0x6969
	call_disconn_ntfy.result_code = 0x04; // call disconnected due to call_clr_rqst
	call_disconn_ntfy.error_code = 0x00;
	call_disconn_ntfy.cause_code = htons(0x0000);
	call_disconn_ntfy.reserved1 = htons(0x0000);
	memcpy(call_disconn_ntfy.call_stats, callStats, 128);
		
	*rply_size = sizeof(call_disconn_ntfy);
	memcpy(rply_packet, &call_disconn_ntfy, *rply_size);
		
	pptpd_debug("CTRL: I've made a suitable CALL DISCONNECT reply packet..", VERBOSE);
			
	return 0;
	
}


/*
int make_in_call_rqst_pckt(int clientFd, unsigned char *rply_packet, int *rply_size) {

	int i;
	
       	struct pptp_header rply_header;
	struct pptp_in_call_rqst in_call_rqst;
	
	unsigned char dialedNumber[64] = {0};
	unsigned char dialingNumber[64] = {0};
	unsigned char subAddress[64] = {0};
	
	
	rply_header.length = htons(0x00dc);
	rply_header.pptp_type = htons(PPTP_CTRL_MESSAGE);
	rply_header.magic = htonl(PPTP_MAGIC_COOKIE);
	rply_header.ctrl_type = htons(IN_CALL_RQST);
	rply_header.reserved0 = htons(0x0000);
		
	in_call_rqst.header = rply_header;
	in_call_rqst.call_id = htons(0x0000);
	in_call_rqst.call_serial = htons(0x0000);
	
	in_call_rqst.bearer_type = htonl(DIGITAL_ACCESS);
	in_call_rqst.channel_id = htonl(0x00000000);
	in_call_rqst.dialed_len = htons(0x0000);
	in_call_rqst.dialing_len = htons(0x0000);
	memcpy(in_call_rqst.dialed_num, dialedNumber, 64);
	memcpy(in_call_rqst.dialing_num, dialingNumber, 64);
	memcpy(in_call_rqst.subaddress, subAddress, 64);
	
	*rply_size = sizeof(in_call_rqst);
	memcpy(rply_packet, &in_call_rqst, *rply_size);
		
	pptpd_debug("I've made a suitable IN CALL REQUEST..", VERBOSE);
			
	return 0;
	
}
*/

